@@ -14,16 +14,24 @@ import time
 
 def check_call(*args, **kwargs):
     """Wraps :func:`subprocess.check_call`."""
-    while True:
-        with contextlib.suppress(subprocess.CalledProcessError):
-            return subprocess.check_call(*args, **kwargs)
+    with open('logs/upload.log', 'wt', buffering=1) as file:
+        if 'stdout' not in kwargs:
+            kwargs['stdout'] = file
+        if 'stderr' not in kwargs:
+            kwargs['stderr'] = subprocess.STDOUT
+
+        for _ in range(3):
+            with contextlib.suppress(subprocess.CalledProcessError):
+                return subprocess.check_call(*args, **kwargs)
+            time.sleep(60)
 
 
 def check_output(*args, **kwargs):
     """Wraps :func:`subprocess.check_output`."""
-    while True:
+    for _ in range(3):
         with contextlib.suppress(subprocess.CalledProcessError):
             return subprocess.check_output(*args, **kwargs)
+        time.sleep(60)
 
 
 def timestamp(container_id):
@@ -64,13 +72,21 @@ def healthcheck(file, interval):
                 # check_call(['docker-compose', '--file', file, 'unpause'])
                 # print(f'[{datetime.datetime.now().isoformat()}] Unpaused container {container_id}')
             if status == 'exited':
+                with open(f'logs/{time.strftime(r"%Y-%m-%d-%H-%M-%S")}.log', 'wb') as log_file:
+                    check_call(['docker', 'logs', '-t', '--details', container_id],
+                               stdout=log_file, stderr=subprocess.STDOUT)
+                check_call(['docker', 'system', 'prune', '--volumes', '-f'])
                 check_call(['docker-compose', '--file', file, 'up', '--detach'])
                 print(f'[{datetime.datetime.now().isoformat()}] Started container {container_id}')
 
             # healthy / unhealthy
             health = info['State']['Health']['Status'].casefold()
             if health == 'unhealthy':
-                check_call(['docker-compose', '--file', file, 'restart'])
+                check_call(['docker', 'stop', container_id])
+                check_call(['docker', 'system', 'prune', '--volumes', '-f'])
+                check_call(['docker-compose', '--file', file, 'up', '--detach'])
+
+                #check_call(['docker-compose', '--file', file, 'restart'])
                 print(f'[{datetime.datetime.now().isoformat()}] Restarted container {container_id}')
 
             # active / inactive
@@ -78,7 +94,14 @@ def healthcheck(file, interval):
             then_ts = timestamp(container_id)
             if last_ts is not None:
                 if then_ts - last_ts < interval:
-                    check_call(['docker-compose', '--file', file, 'restart'])
+                    check_call(['docker', 'stop', container_id])
+                    with open(f'logs/{time.strftime(r"%Y-%m-%d-%H-%M-%S")}.log', 'wb') as log_file:
+                        check_call(['docker', 'logs', '-t', '--details', container_id],
+                                   stdout=log_file, stderr=subprocess.STDOUT)
+                    check_call(['docker', 'system', 'prune', '--volumes', '-f'])
+                    check_call(['docker-compose', '--file', file, 'up', '--detach'])
+
+                    #check_call(['docker-compose', '--file', file, 'restart'])
                     print(f'[{datetime.datetime.now().isoformat()}] Restarted container {container_id}')
             ts_dict[container_id] = then_ts
         time.sleep(interval)
@@ -106,13 +129,24 @@ def main():
     if args.interval <= 0:
         parser.error('invalid interval')
 
-    while True:
-        try:
-            healthcheck(args.file, args.interval)
-        except KeyboardInterrupt:
-            break
-        except Exception:
-            time.sleep(args.interval)
+    os.makedirs('logs', exist_ok=True)
+    if os.path.isfile('logs/healthcheck.log'):
+        os.rename('logs/healthcheck.log', f'logs/healthcheck-{time.strftime(r"%Y%m%d-%H%M%S")}.log')
+
+    with open('logs/healthcheck.log', 'wt', buffering=1) as file:
+        date = datetime.datetime.now().ctime()
+        print('-' * len(date), file=file)
+        print(date, file=file)
+        print('-' * len(date), file=file)
+
+        with contextlib.redirect_stdout(file), contextlib.redirect_stderr(file):
+            while True:
+                try:
+                    healthcheck(args.file, args.interval)
+                except KeyboardInterrupt:
+                    break
+                except Exception:
+                    time.sleep(args.interval)
     return 0
 
 
